"""Singleton behavior implementation."""
# pylint: disable=too-few-public-methods
from typing import Callable
from iocpy.interfaces.instance_behavior import IInstanceBehavior
from iocpy.ioc_context import IocContext


class IocSingleton(IInstanceBehavior):
    """Singleton behavior
        This behavior will create a single instance and return it every time it is resolved.
    :param IInstanceBehavior: _description_
    :type IInstanceBehavior: _type_
    """

    def __init__(self, type_: type, instance: object | Callable[[IocContext], object]):
        self._type = type_
        self._instance = instance

        self._singleton: object | None = None

    def resolve(self, context: IocContext) -> object:
        """Get the instance and follow dependency"""
        root = context.get_root()
        singleton = root.instances.get(self._type, None)
        if singleton is not None:
            return singleton

        if callable(self._instance):
            singleton = self._instance(context)
            root.instances[self._type] = singleton
        else:
            singleton = self._instance
            root.instances[self._type] = singleton
        return singleton



"""IOC Container module."""

from typing import Any, Callable, Generator, Type, TypeVar, ContextManager
from iocpy.interfaces.instance_provider import IInstanceProvider
from iocpy.interfaces.ioc_context import IIocContext
from iocpy.ioc_context import IocContext
from iocpy.ioc_registry import IocRegistry


T = TypeVar("T")


class IocContainer:
    """
    The IOC container is the main entry point for the IOC container.
    This container contains the root context and the behavior registry.

    This is the main public facing API for the IOC container.

    """

    def __init__(self):
        self._registry = IocRegistry()
        self._root_context = IocContext(self._registry)

    def register_singleton(self, type_: type,
                           instance: object | Callable[[IInstanceProvider], object]) -> None:
        """Register a singleton instance in the container.

        The singleton will resolve to the same instance every time it is resolved.

        Example:
        ```PYTHON

        container = IOCContainer()
        container.register_instance(MyInterface, MyImplementation())

        my_instance1 = container.get(MyInterface)
        my_instance2 = container.get(MyInterface)

        assert my_instance1 == my_instance2
        ```

        Args:
            type_ (type): The interface or type that should resolve to the instance
            instance (object | Callable[[IInstanceProvider], object]):
            The instance or a callable that returns the instance
        """
        self._registry.register_singleton(type_, instance)

    def register_instance(self, type_: type, instance: object) -> None:
        """Simple interface to register a singleton instance.

        The singleton will resolve to the same instance every time it is resolved.

        Example:
        ```PYTHON

        container = IOCContainer()
        container.register_instance(MyInterface, MyImplementation())

        my_instance1 = container.get(MyInterface)
        my_instance2 = container.get(MyInterface)

        assert my_instance1 == my_instance2
        ```

        Args:
            type_ (type): The interface or type that should resolve to the instance
            instance (object): The instance to register
        """
        self._registry.register_singleton(type_, instance)

    def register_transient(self, type_: type,
                           instance: Callable[[IInstanceProvider], object]) -> None:
        """ Register a transient instance in the container.

        The transient will created a new instance every time it is resolved.

        Example:
        ```PYTHON

        container = IOCContainer()
        container.register_transient(MyInterface, lambda provider: MyImplementation())

        my_instance = container.get(MyInterface)
        my_instance2 = container.get(MyInterface)

        assert my_instance != my_instance2
        ```


        Args:
            type_ (type): The interface or type that should resolve to the instance
            instance (Callable[[IInstanceProvider], object]):
            The callable that returns the instance
        """
        self._registry.register_transient(type_, instance)

    def register_scoped(self, type_: type,
                        instance: Callable[[IInstanceProvider], object] |
                        Callable[[IInstanceProvider],
                                 Callable[[IInstanceProvider], Generator[Any, None, None]]] |
                        Callable[[IInstanceProvider],
                                 Generator[Any, None, None]]
                        ) -> None:
        """
        Register a scoped instance in the container.

        The scoped instance will be created once per scope.
        This means inside the open scope the type will always resolve to the same
        instance.
        example:
        ```PYTHON	

        container = IOCContainer()
        container.register_scoped(MyInterface, lambda provider: MyImplementation())
        container.register_scoped(MyInterfaceWrapper, 
            lambda provider: MyWrapper(provider.get(MyInterface))
        )
        with container.create_scope() as scope:
            my_instance = scope.get(MyInterface)
        ```

        Args:
            type_ (type): The interface or type that should resolve to the instance
            instance (
            Callable[[IInstanceProvider], object] |
            Callable[[IInstanceProvider], Callable[[IInstanceProvider], 
            Generator[Any, None, None]]] |
              Callable[[IInstanceProvider], Generator[Any, None, None]]
              ): The callable that returns the instance
        """
        self._registry.register_scoped(type_, instance)

    def get(self, type_: Type[T]) -> T:
        """Resolve the instance for the given type.

        Args:
            type_ (Type[T]): The interface or class to resolve.

        Returns:
            T: The resolved instance `T` for the given type.
        """
        return self._root_context.get(type_)

    def create_scope(self) -> ContextManager[IIocContext]:
        """Create a new scope using a context manager.

        Returns:
            ContextManager[IIocContext]: The contextmanager
        """
        return self._root_context.create_scope()

    def __call__(self) -> Generator[IIocContext, None, None]:
        # ? We need to call it this way that the function signature is a callable generator
        # pylint: disable=not-context-manager
        """Callable generator scope, can be used with FastAPI dependency injection
        Example:
        ```PYTHON

        IOC = IocContainer()

        app.get("/items/{item_id}")
        async def read_item(item_id: int, scope: IIocContext = Depends(IOC)):
            item = scope.get(ItemService).get_item(item_id)
            return item
        ```
        """
        with self.create_scope() as context:
            yield context



"""Transient behavior implementation."""
# pylint: disable=too-few-public-methods
from typing import Callable
from iocpy.interfaces.instance_behavior import IInstanceBehavior
from iocpy.ioc_context import IocContext


class IocTransient(IInstanceBehavior):
    """Transient behavior
        This behavior will create a new instance every time it is resolved.
    """

    def __init__(self, type_: type, instance: Callable[[IocContext], object]):
        self._type = type_
        self._instance = instance

    def resolve(self, context: IocContext) -> object:
        """Get the instance and follow dependency"""
        return self._instance(context)



"""
Module for the IocContext class implementing IIocContext and IInstanceProvider.
"""

from contextlib import ExitStack, contextmanager
import logging
from typing import Any, Generator, Type, TypeVar
from iocpy.interfaces.behavior_registry import IBehaviorRegistry
from iocpy.interfaces.instance_provider import IInstanceProvider
from iocpy.interfaces.ioc_context import IIocContext


T = TypeVar("T")


class IocContext(IInstanceProvider, IIocContext):
    """The IocContext is used for the implementation of scopes.
    Every scope represents a new context that can be used to resolve instances.

    The newly resolved instances will be stored in the context and will be 
    cleaned up when the context is closed.
    The parent argument is used to recursively get the root context.


    Args:
        IInstanceProvider (_type_): 
        Minimal interface to resolve instances with `.get(<type>)`
        IIocContext (_type_): 
        The context interface is the public facing interface to the IOC container.
    """

    def __init__(self, registry: IBehaviorRegistry,
                 parent: IIocContext | None = None, stack: ExitStack | None = None):
        self._registry = registry
        self._parent = parent
        self._stack: ExitStack | None = stack
        self._instances: dict[Type, object] = {}
        self._logger = logging.getLogger(__name__)

    def get_root(self) -> IIocContext:
        """Get the root context throng recursion

        Returns:
            IIocContext: The root context.
        """
        if self._parent is None:
            return self
        return self._parent.get_root()

    @contextmanager
    def create_scope(self) -> Generator[IIocContext, Any, None]:
        """Create a new scope with this scope as its parent.

        Returns:
            ContextManager[IIocContext]: 
            Context manager containing the new scope.
_
        """
        return self.create_generator_scope()

    def create_generator_scope(self) -> Generator[IIocContext, None, None]:
        """Create a new generator scope.

        Yields:
            Generator[IIocContext, None, None]: The generator scope.
        """
        with ExitStack() as stack:
            context = IocContext(self._registry, self, stack)
            try:
                yield context
            finally:
                self._logger.debug("Leaving scope, cleaning up")
        self._instances.clear()

    def get(self, type_: Type[T]) -> T:
        """Resolve a instance from the container.

        Args:
            type_ (Type[T]): The interface or class to resolve.

        Raises:
            ValueError: Invalid resolve type error.

        Returns:
            T: The resolved instance `T` for the given type.
        """
        self._logger.debug("Getting instance of %s", type_)
        behavior = self._registry.get_behavior(type_)
        instance = behavior.resolve(self)

        if not isinstance(instance, type_):
            raise ValueError(f"The resolved instance is not of type {type_}")
        return instance

    @property
    def stack(self) -> ExitStack | None:
        """The exit stack for the scoped dependencies.
        This stack is used to manage the lifecycle of the scoped dependencies.

        Returns:
            ExitStack | None: The exit stack for the scoped dependencies.
        """
        return self._stack

    @property
    def instances(self) -> dict[Type, object]:
        """Registered scoped instances for this scope

        Returns:
            dict[Type, object]: 
            Resolved and registered scoped instances.
        """
        return self._instances




from typing import Generator
from unittest import TestCase
from unittest.mock import MagicMock

from iocpy.ioc_container import IocContainer


class TestIocContainer(TestCase):

    def test_RegisterSingleton_ReturnsSameInstance(self):
        # Arrange
        container = IocContainer()
        call_spy = MagicMock(side_effect=[1, 2, 3, 4])
        container.register_singleton(int, lambda x: call_spy())

        # Act
        instance1 = container.get(int)
        instance2 = container.get(int)

        # Assert
        self.assertEqual(instance1, instance2)
        self.assertEqual(call_spy.call_count, 1)

    def test_RegisterTransient_ReturnsDifferentInstances(self):
        # Arrange
        container = IocContainer()
        call_spy = MagicMock(side_effect=[1, 2, 3, 4])
        container.register_transient(int, lambda x: call_spy())

        # Act
        instance1 = container.get(int)
        instance2 = container.get(int)

        # Assert
        self.assertNotEqual(instance1, instance2)
        self.assertEqual(call_spy.call_count, 2)

    def test_RegisterScoped_ReturnsSameInstancePerScope(self):
        # Arrange
        container = IocContainer()
        call_spy = MagicMock(side_effect=[1, 2, 3, 4])
        container.register_scoped(int, lambda x: call_spy())

        # Act & Assert
        with container.create_scope() as scope:
            self.assertEqual(scope.get(int), 1)
            self.assertEqual(scope.get(int), 1)

        with container.create_scope() as scope:
            self.assertEqual(scope.get(int), 2)
            self.assertEqual(scope.get(int), 2)

        # Assert
        self.assertEqual(call_spy.call_count, 2)

    def test_RegisterDependentScope_ResolvesDependant(self):
        # Arrange
        class TestImplementation:
            def get_session(self, *args, ** kwargs) -> Generator[int, None, None]:
                yield 1
        container = IocContainer()
        container.register_instance(object, TestImplementation())
        container.register_scoped(int, lambda x: x.get(object).get_session)

        # Act & Assert
        with container.create_scope() as scope:
            self.assertEqual(scope.get(int), 1)

    def test_ResolveScopedInRootScope_ThrowsScopeError(self):
        # Arrange
        container = IocContainer()
        container.register_scoped(int, lambda x: 1)

        # Act & Assert
        with self.assertRaises(ValueError) as ex:
            container.get(int)

        self.assertEqual(str(ex.exception),
                         "Scoped instances can't be resolved outside of a scope")

    def test_ScopedSession_ClosedOnScopeExit(self):
        # Arrange
        container = IocContainer()
        enter = MagicMock()
        exit = MagicMock()
        instance = MagicMock(side_effect=[1, 2])

        def create_session(context):
            enter()
            yield instance()
            exit()
        container.register_scoped(int, create_session)

        # Act
        with container.create_scope() as scope:
            instance1 = scope.get(int)
            instance2 = scope.get(int)

        # Assert
        self.assertEqual(instance1, instance2)
        enter.assert_called_once()
        exit.assert_called_once()

    def test_RegisterInstances_ThrowsExistsError(self):
        # Arrange
        container = IocContainer()
        container.register_instance(int, 1)

        # Act & Assert
        with self.assertRaises(ValueError) as ex:
            container.register_instance(int, 1)

        self.assertEqual(str(ex.exception),
                         "Type <class 'int'> already registered")

    def test_RetrieveInstance_ThrowsNotExistsError(self):
        # Arrange
        container = IocContainer()

        # Act & Assert
        with self.assertRaises(ValueError) as ex:
            container.get(int)

        self.assertEqual(str(ex.exception),
                         "Type <class 'int'> not registered")

    def test_RetrieveWithIncompatibleTypeInstance_ThrowsError(self):
        # Arrange
        container = IocContainer()
        container.register_instance(int, "abc")

        # Act & Assert
        with self.assertRaises(ValueError) as ex:
            container.get(int)

        self.assertEqual(str(ex.exception),
                         "The resolved instance is not of type <class 'int'>")



"""instance behavior interface."""
# pylint: disable=too-few-public-methods
from abc import ABC, abstractmethod
from typing import Generic, TypeVar, TYPE_CHECKING

if TYPE_CHECKING:
    from iocpy.ioc_context import IocContext


T = TypeVar("T")


class IInstanceBehavior(ABC, Generic[T]):
    """
    The behavior interface is used to resolve the instance
    in different ways like singleton, transient, etc.
    """

    @abstractmethod
    def resolve(self, context: 'IocContext') -> T:
        """Resolves a instance, and recuse over dependency's"""



"""
This module contains the interface for the behavior registry.
The behavior registry is used to register different types of behaviors
like singleton, transient, etc.
This is used then used by the IIOContext to resolve instances.
"""

from abc import ABC, abstractmethod
from typing import Callable

from iocpy.interfaces.instance_behavior import IInstanceBehavior
from iocpy.interfaces.instance_provider import IInstanceProvider


class IBehaviorRegistry(ABC):
    """
    The behavior registry is used to register different types of behaviors
    """

    @abstractmethod
    def get_behavior(self, type_: type) -> IInstanceBehavior:
        """Get the behavior for a type.
        The behavior is warping the registered callables and instances.
        """

    @abstractmethod
    def register_singleton(self, type_: type,
                           instance: object | Callable[[IInstanceProvider], object]) -> None:
        """
        Register a singleton instance.

        Args:
            type_ (type): The interface or class to register for.
            instance (object | Callable[[IInstanceProvider], object]): 
            The instance or callable to resolve the type for.
        """

    @abstractmethod
    def register_transient(self, type_: type,
                           instance: Callable[[IInstanceProvider], object]) -> None:
        """
        Register a transient.

        Args:
            type_ (type): The interface or class to register for. 
            instance (Callable[[IInstanceProvider], object]):
            The callable to resolve the type for.
        """

    @abstractmethod
    def register_scoped(self, type_: type, instance: Callable[[IInstanceProvider], object]) -> None:
        """
        Register a scoped instance.

        Args:
            type_ (type): The interface or class to register for.
            instance (Callable[[IInstanceProvider], object]):
            The callable to resolve the type for.
        """



"""interfaces/instance_provider.py"""
# pylint: disable=too-few-public-methods
from abc import ABC, abstractmethod
from typing import Type, TypeVar

T = TypeVar("T")


class IInstanceProvider(ABC):
    """Instance provider interface, provides instances of the specified type."""

    @abstractmethod
    def get(self, type_: Type[T]) -> T:
        """Get an instance of the specified type."""



"""IOC Context interface module.
The context interface is the public facing interface to the IOC container.
"""

from abc import ABC, abstractmethod
from contextlib import ExitStack
from typing import Generator, Type, TypeVar, ContextManager


T = TypeVar("T")


class IIocContext(ABC):
    """
    The context interface is the public facing 
    interface to the IOC container.
    """

    @abstractmethod
    def get_root(self) -> 'IIocContext':
        """Get the root context.
        The root context contains all singleton instances.

        Returns:
            IIocContext: The root context.
        """

    @abstractmethod
    def create_scope(self) -> ContextManager['IIocContext']:
        """Create a scope using a context manager.
        The scope will be cleaned up after the context manager is done.
        """

    @abstractmethod
    def create_generator_scope(self) -> Generator['IIocContext', None, None]:
        """Create a new generator scope.
        This scope type can be used to integrate with other frameworks
        or in combination with the context manager decorator.
        """

    @abstractmethod
    def get(self, type_: Type[T]) -> T:
        """
        Resolve a instance from the registry.

        Args:
            type_ (Type[T]): The interface or class to resolve.

        Returns:
            T: The resolved instance.
        """

    @property
    @abstractmethod
    def stack(self) -> ExitStack | None:
        """Exit stack for scoped dependencies."""

    @property
    @abstractmethod
    def instances(self) -> dict[Type, object]:
        """Instance dictionary for the (scoped) context."""



"""Scoped behavior implementation"""
from contextlib import ExitStack, contextmanager
import inspect
from typing import Any, Callable, Generator
from iocpy.interfaces.instance_behavior import IInstanceBehavior
from iocpy.ioc_context import IocContext


class IocScoped(IInstanceBehavior):
    """Scoped behavior
        This behavior will create only one instance per scope / context.
    """

    def __init__(self, type_: type, factory:
                 Callable[[IocContext], object] |
                 Callable[[IocContext], Callable[[IocContext], Generator[Any, None, None]]] |
                 Callable[[IocContext], Generator[Any, None, None]]
                 ):
        self._type = type_
        self._factory = factory

    def resolve(self, context: IocContext) -> object:
        """Resolve the instance and follow dependencies

        Args:
            context (IocContext): The context used to resolve the instance.

        Raises:
            ValueError: Invalid factory / callable type
            ValueError: Scoped instances can't be resolved outside of a scope

        Returns:
            object: The resolved instance.
        """

        if context.stack is None:
            raise ValueError(
                "Scoped instances can't be resolved outside of a scope")

        scoped = context.instances.get(self._type, None)
        if scoped is not None:
            return scoped

        # TODO: We most likely should cache this in the future for performance reasons
        if inspect.isgeneratorfunction(self._factory):
            scoped = self.handle_generator(
                context, self._factory, context.stack)
            context.instances[self._type] = scoped
            return scoped

        if callable(self._factory):
            scoped = self._factory(context)
            if inspect.isgeneratorfunction(scoped):
                scoped = self.handle_generator(
                    context, scoped, context.stack)
            context.instances[self._type] = scoped
            return scoped

        raise ValueError("Failed to resolve, invalid factory")

    def handle_generator(self, context: IocContext,
                         factory: Callable[[IocContext],
                                           Generator[Any, None, None]],
                         stack: ExitStack
                         ) -> object:
        """Get instance from generator function and 
        add it to the context exit stack

        Args:
            context (IocContext): The context to resolve nested dependencies.
            factory (Callable[[IocContext], Generator[Any, None, None]]): 
            The generator function to resolve.
            stack (ExitStack): The exit stack to add the generator to.

        Returns:
            object: The instance resolved from the generator.
        """
        cm = contextmanager(factory)(context)
        scoped = stack.enter_context(cm)
        return scoped



"""This module implements the IIocRegistry interface."""
from typing import Any, Callable, Generator
from iocpy.behaviors.scoped import IocScoped
from iocpy.behaviors.singleton import IocSingleton
from iocpy.behaviors.transient import IocTransient
from iocpy.interfaces.behavior_registry import IBehaviorRegistry
from iocpy.interfaces.instance_behavior import IInstanceBehavior
from iocpy.interfaces.instance_provider import IInstanceProvider


class IocRegistry(IBehaviorRegistry):
    """
    The ioc registry contains all behavior like 
    singleton, transient, and scoped instances.
    The registry in mainly used by the `IocContext` to resolve instances.

    The class might change and is not a public API.
    Use the `IocContainer` to register instances that wraps this registry.

    Args:
        IBehaviorRegistry (_type_): 
        The behavior registry is used to register different types of behaviors
    """

    def __init__(self):
        self._registry: dict[type, IInstanceBehavior] = {}

    def register(self, type_: type, instance: IInstanceBehavior) -> None:
        """Register a custom instance behavior implementing `IInstanceBehavior`.

        Args:
            type_ (type): The type that should resolve to the instance
            instance (IInstanceBehavior): The instance behavior

        Raises:
            ValueError: The type is already registered.
        """
        if type_ in self._registry:
            raise ValueError(f"Type {type_} already registered")
        self._registry[type_] = instance

    def get_behavior(self, type_: type) -> IInstanceBehavior:
        """Get the instance behavior for a type.

        Args:
            type_ (type): The type to get the behavior for.

        Raises:
            ValueError: The type is not registered.

        Returns:
            IInstanceBehavior: The instance behavior for the type.
        """
        if type_ not in self._registry:
            raise ValueError(f"Type {type_} not registered")
        return self._registry[type_]

    def register_singleton(self, type_: type,
                           instance: object | Callable[[IInstanceProvider], object]) -> None:
        """Register a singleton instance in the container.

        Example:
        ```PYTHON
        container = IOCContainer()
        container.register_singleton(MyInterface, MyImplementation())
        container.register_singleton(MyInterfaceWrapper, 
            lambda provider: MyWrapper(provider.get(MyInterface))
        )

        my_instance = container.get(MyInterface)
        my_instance_wrapper = container.get(MyInterfaceWrapper)

        ```
        :param type_: The type (this can be a interface or a class)
        :type type_: type
        :param instance: The instance or a callable that returns the instance
        :type instance: object | Callable[[IInstanceProvider], object]
        """
        singleton = IocSingleton(type_, instance)
        self.register(type_, singleton)

    def register_transient(self, type_: type,
                           instance: Callable[[IInstanceProvider], object]) -> None:
        """Register a transient instance in the container.
        This will create a new instance every time `get` is called.

        Example:
        ```PYTHON
        container = IOCContainer()
        container.register_transient(MyInterface, lambda provider: MyImplementation())

        my_instance = container.get(MyInterface)

        :param type_: _description_
        :type type_: type
        :param instance: _description_
        :type instance: Callable[[IInstanceProvider], object]
        """
        transient = IocTransient(type_, instance)
        self.register(type_, transient)

    def register_scoped(self, type_: type, instance:
                        Callable[[IInstanceProvider], object] |
                        Callable[[IInstanceProvider],
                                 Callable[[IInstanceProvider], Generator[Any, None, None]]] |
                        Callable[[IInstanceProvider],
                                 Generator[Any, None, None]]
                        ) -> None:
        """Register a scoped instance in the container.
        This will create only one instance per scope.

        Errors will be raised if the instance is resolved outside of a scope.

        Example:
        ```PYTHON
        container = IOCContainer()
        container.register_scoped(MyInterface, lambda provider: MyImplementation())

        my_instance = container.get(MyInterface)

        ```
        show in to the `README.md` file for more complex examples with sessions and generators.

        :param type_: _description_
        :type type_: type
        :param instance: _description_
        :type instance: Callable[[IocContext], object]
        """
        scoped = IocScoped(type_, instance)
        self.register(type_, scoped)
